﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Resources;
using System.Threading;
using System.Reflection;
using System.IO;
using System.Globalization;
using System.Diagnostics;
using RootPro.RootProCAD;
using RootPro.RootProCAD.Command;
using RootPro.RootProCAD.Geometry;
using RootPro.RootProCAD.UI;

//////////////////////////////////////////////////////////////////
// 分割割付のアドイン
// 2 つの図形の間に、等間隔に図形を作成します。
//////////////////////////////////////////////////////////////////

namespace RCAddInAllocationCS
{
    [System.AddIn.AddIn("AppAddIn", Version = "1.0", Publisher = "", Description = "")]
    public partial class AppAddIn
    {
        string AllocationCommandName = "RootPro.AddIn.AllocationCommand";     //コマンド名 
        
        private void AppAddIn_Startup(object sender, EventArgs e)
        {
			AppDomain.CurrentDomain.AssemblyResolve += new ResolveEventHandler(ResolveAssembly);

			CommandManager.AddMacroCommand("RCAddInAllocationCS", Properties.Resources.CommandName, CreateAllocationShape);

            //結合コマンドを追加する
            Command command = CommandManager.AddCommand(AllocationCommandName, Properties.Resources.CommandName);

            //パラメータを追加する
            for (int i = 0; i < 2; i++)
            {
                CommandParameterGroup commandParameterGroup = command.ParameterGroups.Add(ParameterItemType.Shape, i, Properties.Resources.ShapeParameterName + (i + 1));
                ShapeParameterItem parameterItem = (ShapeParameterItem)commandParameterGroup.CurrentItem; // パラメータアイテム
                parameterItem.ShapeTypeFilter = ShapeTypeFilters.Line | ShapeTypeFilters.Circle;      //線分と連続線のみ
                parameterItem.Description = Properties.Resources.ShapeParameterDescription;
                parameterItem.ValueSaveMode = ParameterValueSaveMode.OneTime;
            }

            command.ParameterGroups.Add(ParameterItemType.Integer, 2, Properties.Resources.DivisionNumberParameterName);
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetParameterItemByID(2);
            integerParameterItem.Description = Properties.Resources.DivisionNumberParameterDescription;
            integerParameterItem.Value = 2;
            integerParameterItem.ValueSaveMode = ParameterValueSaveMode.Always;

            //開始イベントと終了イベントの追加
            command.Begin += new CommandBeginEventHandler(command_Begin);
            command.End += new EventHandler(command_End);
        }

        private void AppAddIn_Shutdown(object sender, EventArgs e)
        {
            CommandManager.RemoveMacroCommand(CreateAllocationShape);
            CommandManager.RemoveCommand(AllocationCommandName);
        }

        #region VSTA generated code
        private void InternalStartup()
        {
            this.Startup += new System.EventHandler(AppAddIn_Startup);
            this.Shutdown += new System.EventHandler(AppAddIn_Shutdown);
        }
        #endregion

        /// <summary>
        /// 分割割付コマンド処理
        /// </summary>
        private void CreateAllocationShape()
        {
            //コマンドの実行
            CommandManager.ExecuteCommand(AllocationCommandName);
        }

        /// <summary>
        /// コマンドの開始時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_Begin(Object sender, CommandBeginEventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの追加
            command.ParameterChanged += new CommandParameterChangedEventHandler(command_ParameterChanged);
        }

        /// <summary>
        /// コマンドの終了時に呼ばれる
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_End(Object sender, EventArgs e)
        {
            Command command = (Command)sender;

            //パラメータ変更イベントの削除
            command.ParameterChanged -= new CommandParameterChangedEventHandler(command_ParameterChanged);
        }

        /// <summary>
        /// 分割割付を実行する
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void command_ParameterChanged(Object sender, CommandParameterChangedEventArgs e)
        {
            Document doc = ActiveDocument;         //ドキュメントの取得

            Command command = (Command)sender;
            int id = e.ParameterID;// 変更のあったパラメータID
            ParameterItem parameterItem = command.GetParameterItemByID(id); // パラメータアイテム

            //図形１の取得
            ShapeParameterItem shapeParameterItem1 = (ShapeParameterItem)command.GetParameterItemByID(0);
            ShapeParameterItem shapeParameterItem2 = (ShapeParameterItem)command.GetParameterItemByID(1);
            if (shapeParameterItem1.IsEmpty)
            {
                if (!shapeParameterItem2.IsEmpty)
                {
                    Shape shape2a = shapeParameterItem2.Shape.Shape;
                    if (shape2a is LineShape)
                        shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Line;
                    else
                        shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Circle;
                }
                else
                {
                    shapeParameterItem1.ShapeTypeFilter = ShapeTypeFilters.Circle | ShapeTypeFilters.Line;
                    shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Circle | ShapeTypeFilters.Line;
                }
                return;
            }
            Shape shape1 = shapeParameterItem1.Shape.Shape;
            Point2d pickPoint1 = shapeParameterItem1.PickPoint;

            //図形２の取得
            if (shapeParameterItem2.IsEmpty)
            {
                if (shape1 is LineShape)
                    shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Line;
                else
                    shapeParameterItem2.ShapeTypeFilter = ShapeTypeFilters.Circle;
                return;
            }
            Shape shape2 = shapeParameterItem2.Shape.Shape;
            Point2d pickPoint2 = shapeParameterItem2.PickPoint;

            //図形が同じタイプか？
            if (shape1.GetType() != shape2.GetType())
            {
                MessageBox.Show(Properties.Resources.ConfirmMessage1);
                return;
            }

            //分割数の取得
            IntegerParameterItem integerParameterItem = (IntegerParameterItem)command.GetParameterItemByID(2);
            if (integerParameterItem.IsEmpty)
                return;
            int dividCount = integerParameterItem.Value;
            if (dividCount < 2 || dividCount > 100)
            {
                MessageBox.Show(Properties.Resources.ConfirmMessage2);
                return;
            }

            // 用紙ビューの場合
            DrawingView drawingView = (DrawingView)ActiveTabView;
            if (drawingView.Drawing is Paper)
            {
                // 図形を用紙の座標系に変換する
                shape1 = drawingView.TransformShapeLocalToWorld(shapeParameterItem1.Shape.Layout, shape1);
                shape2 = drawingView.TransformShapeLocalToWorld(shapeParameterItem2.Shape.Layout, shape2);
            }

            //分割図形の作成
            Shape[] shapes = CreateAllocationShape(shape1, shape2, pickPoint1, pickPoint2, dividCount);
            if(shapes == null || shapes.Length == 0)
            {
                e.CommandState = CommandState.CommandError;
                return;
            }

            // UNDO 可能な操作を開始
            doc.UndoManager.BeginUndoUnit();

            foreach(Shape shape in shapes)
            {
                Shape copyShape = shape.Clone();

                copyShape.Layer = doc.LayerTable.CurrentLayer;
                doc.BasicShapeSettings.CopyToShape(copyShape);

                // 用紙ビューの場合、カレント部分図の座標系に変換する
                if (drawingView.Drawing is Paper)
                    copyShape = drawingView.TransformShapeWorldToLocal(ActiveDocument.CurrentLayout, copyShape);

                doc.CurrentDrawing.Shapes.Add(copyShape);
            }

            // UNDO 可能な操作を終了
            doc.UndoManager.EndUndoUnit();

            e.CommandState = CommandState.Execute;
        }

        /// <summary>
        /// 分割割付の実行
        /// </summary>
        /// <param name="shape1"></param>
        /// <param name="shape2"></param>
        /// <param name="pickPoint1"></param>
        /// <param name="pickPoint2"></param>
        /// <param name="dividCount"></param>
        private Shape[] CreateAllocationShape(Shape shape1, Shape shape2, Point2d pickPoint1, Point2d pickPoint2, int dividCount)
        {
            List<Shape> shapeList = new List<Shape>();

            Type type = shape1.GetType();
            if (type.Name == "CircleShapeClass")        //円の分割割付
            {
                CircleShape circle1 = (CircleShape)shape1;
                CircleShape circle2 = (CircleShape)shape2;

                for (int i = 0; i < dividCount - 1; i++)
                {
                    double cx = circle1.Center.X + (circle2.Center.X - circle1.Center.X) * (i + 1) / dividCount;
                    double cy = circle1.Center.Y + (circle2.Center.Y - circle1.Center.Y) * (i + 1) / dividCount;
                    double radius = circle1.Radius + (circle2.Radius - circle1.Radius) * (i + 1) / dividCount;

                    Point2d centerPoint = Geometry.CreatePoint(cx, cy);

                    shapeList.Add(ShapeFactory.CreateCircle(centerPoint, radius));
                }
            }
            else if (type.Name == "LineShapeClass")     //線分の分割割付
            {
                LineShape line1 = (LineShape)shape1;
                LineShape line2 = (LineShape)shape2;

                double sx1 = line1.StartPoint.X;
                double sy1 = line1.StartPoint.Y;
                double ex1 = line1.EndPoint.X;
                double ey1 = line1.EndPoint.Y;

                double sx2 = line2.StartPoint.X;
                double sy2 = line2.StartPoint.Y;
                double ex2 = line2.EndPoint.X;
                double ey2 = line2.EndPoint.Y;

                InfiniteLineShape infiniteLineShape1 = ShapeFactory.CreateInfiniteLine(line1.StartPoint, line1.Angle);
                InfiniteLineShape infiniteLineShape2 = ShapeFactory.CreateInfiniteLine(line2.StartPoint, line2.Angle);
                IntersectionPoint[] intersectionPoints = infiniteLineShape1.GetIntersectionPoint(infiniteLineShape2);
                if(intersectionPoints != null && intersectionPoints.Length > 0)
                {
                    Point2d intersectPoint = intersectionPoints[0].PointOnShape1.Point;

                    PointOnShape pointOnShape1 = line1.GetClosestPoint(pickPoint1);
                    if (pointOnShape1 != null)
                    {
                        double lineAngle = Geometry.DegreeToRadian(line1.Angle);
                        double angle = Geometry.DegreeToRadian(Geometry.GetAngle(intersectPoint, pointOnShape1.Point));
                        if (!Geometry.IsRadianEquals(lineAngle, angle))
                        {
                            sx1 = line1.EndPoint.X;
                            sy1 = line1.EndPoint.Y;
                            ex1 = line1.StartPoint.X;
                            ey1 = line1.StartPoint.Y;
                        }
                    }

                    PointOnShape pointOnShape2 = line2.GetClosestPoint(pickPoint2);
                    if (pointOnShape2 != null)
                    {
                        double lineAngle = Geometry.DegreeToRadian(line2.Angle);
                        double angle = Geometry.DegreeToRadian(Geometry.GetAngle(intersectPoint, pointOnShape2.Point));
                        if (!Geometry.IsRadianEquals(lineAngle, angle))
                        {
                            sx2 = line2.EndPoint.X;
                            sy2 = line2.EndPoint.Y;
                            ex2 = line2.StartPoint.X;
                            ey2 = line2.StartPoint.Y;
                        }
                    }
                }

                for (int i = 0; i < dividCount - 1; i++)
                {
                    double x1 = sx1 + (sx2 - sx1) * (i + 1) / dividCount;
                    double y1 = sy1 + (sy2 - sy1) * (i + 1) / dividCount;
                    double x2 = ex1 + (ex2 - ex1) * (i + 1) / dividCount;
                    double y2 = ey1 + (ey2 - ey1) * (i + 1) / dividCount;

                    Point2d point1 = Geometry.CreatePoint(x1, y1);
                    Point2d point2 = Geometry.CreatePoint(x2, y2);

                    shapeList.Add(ShapeFactory.CreateLine(point1, point2));
                }
            }

            return shapeList.ToArray();
        }

		private Assembly ResolveAssembly(object sender, ResolveEventArgs args)
		{
			AssemblyName assemblyName = new AssemblyName(args.Name);

			foreach (Assembly assembly in AppDomain.CurrentDomain.GetAssemblies())
			{
				if (AssemblyName.ReferenceMatchesDefinition(assemblyName, assembly.GetName()))
					return assembly;
			}

			if (assemblyName.Name.EndsWith(".resources"))
			{
				if (assemblyName.CultureName.Equals(Thread.CurrentThread.CurrentUICulture.Name, StringComparison.OrdinalIgnoreCase))
				{
					string dllPath = Path.GetDirectoryName(this.AddInFullPath);
					dllPath = Path.Combine(dllPath, assemblyName.CultureName);
					dllPath = Path.Combine(dllPath, assemblyName.Name + ".dll");

					if (File.Exists(dllPath))
					{
						try
						{
							return Assembly.LoadFrom(dllPath);
						}
						catch
						{
						}
					}
				}
			}

			return null;
		}
	}
}
